/* 
  Prctica de PS Quadrimestre Tardor 2004-2005 <EDAMAZE>
  Autors:
          Daniel Clemente Laboreo <www.danielclemente.com>
          Francesc Xavier Morales Pascual <www.xavimorales.biz>
*/

// dedalus.cpp: construcci de laberints perfectes


#include "particio.hpp"
#include "dedalus.hpp"

/************ COMENTARI DE LA FUNCI CONSTRUIR ***********************
  Per aconseguir l'excavaci del laberint fins a obtenir-ne un de perfecte usem l'algorisme de Kruskal, 
  el qual serveix realment per obtenir un "arbre de recorregut mnim" per a un graf. Aquest algorisme s 
  apropiat per al que volem fer perqu uneix els vertex del graf de forma que no es formin cicles i cerca 
  el cam que t un cost ms baix. Aquesta ltima part no ens interesa molt ja que tenim cost igual per a
  totes les portes, per s molt til per a aconseguir un laberint perfecte.
  L'algorisme consisteix en tenir un bosc d'arbres, inicialment de n arbres (tants com vertex), i que a cada pas
  n'uneix dos d'aquests arbres, de forma que aquests queden units en el mateix grup. En cas que estiguessin 
  en el mateix arbre seria innecessari agrupar-los i per tant no els uneix evitant cicles.
  Bsicament l'algorisme funciona aix:
    1. Es construeix el bosc amb cada node en un arbre separat
    2. Fins que no hem afegit n-1 vertex agafem un de la cua, el que t menys cost. Si forma un cicle el rebutjem i sin llavors l'afegim
    
  L'algorisme acaba quan ja tenim tots els nodes units a UN sol arbre.
  
  Podem aplicar aquest sistema en el nostre laberint. El nostre bosc ser una partici de cambres, 
  per per poder-les distingir unes de les altres ens basarem en la seva posici al laberint, creant
  una partici de posicions. L'iniciarem amb totes les posicions formant grups independents i de forma 
  aleatoria buscarem una cambra i un sentit en el que obrir la porta. Mirarem si les dues cambres 
  adjacents formen part del mateix grup i en cas negatiu unirem els dos grups a la vegada que obrim 
  una porta al laberint entre les cambres. Fet aix, repetim el procs fins que ens queda un sol grup.
***********************************************************************/

void dedalus::construir(laberint& M) throw(error) {
  // Construim un laberint perfecte si no est encara excavat
  nat i, j;
  nat nfiles, ncolum;
  cambra tancada;
  posicio actual, contigu;

  // creem una partici de posicions amb la que treballem per saber si podem unir dues cambres i si hem acabat
  particio<posicio> parts;
  util::Random r;
  paret p;
  tancada = cambra(0,0,0,0); // la utilitzarem per comparar les cambres del laberint amb la que t totes les portes tancades (aquesta)

  nfiles = M.num_files();
  ncolum = M.num_columnes();

  //comprovem que totes les cambres estan tancades, recordem que la primera posici s la <1,1>
  for(i=1;i<=nfiles;++i) {
    for(j=1;j<=ncolum;++j) {
      if(tancada!=M(make_pair(i,j)))
        throw error(nom_mod, EstaExcavat, MsgEstaExcavat);
    }
  }

  // afegim cada cambra (representada per la seva posici) com a un grup separat de la partici
  for(i=1;i<=nfiles;++i) {
    for(j=1;j<=ncolum;++j) {
      parts.afegir(make_pair(i,j));

    }
  }

  // mentre tinguem ms d'un grup a la partici hem de seguir unint, perqu significa que encara no hem unit totes les cambres
  while (parts.size() >1) {

    //obtenim un valor aleatori de fila i un altre de columna
    i = (nat)((double)nfiles*r())+1;
    j = (nat)((double)ncolum*r())+1;
    actual = make_pair(i,j);

    //obtenim una paret aleatria
    p = paret((int)(4*r()));

    //calculem els ndex de la cambra adjacent a aquesta paret
    switch((int)p) {
    case paret::NORD:
      --i;
      break;
    case paret::EST:
      ++j;
      break;
    case paret::SUD:
      ++i;
      break;
    case paret::OEST:
      --j;
      break;
    }


    // Si la cambra contgua est fora del laberint (hem intentat obrir una paret exterior) provem amb una nova cambra
    if( i<1 || i>nfiles || j<1 || j>ncolum )
      continue;

    contigu = make_pair(i,j);

    //Si tenen representants diferents significa que podem obrir la porta sense por a crear un cicle
    if (parts.representant(actual)!=parts.representant(contigu)) {
      parts.unir(actual,contigu);
      M.obre_porta(p,actual);
    }
  } // Fi del while

  // Laberint excavat i perfecte
}
